//
// Ethernet + Wifi to WiFi test
//


#include "Arduino.h"
#include <atomic>

#define DEBUG   // enable debug to DebugPort serial port

#ifdef DEBUG
#define DebugPort Serial    // Programming port
#define DBPrint(x) if(DebugPort) DebugPort.print(x)
#define DBPrintln(x) if(DebugPort) DebugPort.println(x)
#define DBPrintHex(x) if(DebugPort) DebugPort.print(x, HEX)
#else
#define DBPrint(x)
#define DBPrintln(x)
#define DBPrintHex(x)
#endif // DEBUG

#define WIFI_VAR_LEN 64

typedef struct IPCONFIG {
	bool            bUseDHCP;
	IPAddress       ip;
	IPAddress       dns;
	IPAddress       gateway;
	IPAddress       subnet;
} IPConfig;

typedef struct WIFICONFIG {
	IPAddress       ip;
	char 			sSSID[WIFI_VAR_LEN];
	char			sPassword[WIFI_VAR_LEN];
} WIFIConfig;

String IpAddress2String(const IPAddress& ipAddress)
{
  return String(ipAddress[0]) + String(".") +
  		String(ipAddress[1]) + String(".") +
		String(ipAddress[2]) + String(".") +
		String(ipAddress[3]); 
}

// includes and some defines for ethernet connection
#include <SPI.h>    // RP2040 :  SCK: GP18, COPI/TX: GP19, CIPO/RX: GP16, CS: GP17
#include <W5500lwIP.h>
#define ETHERNET_CS     17
#define ETHERNET_INT	27
#define ETHERNET_RESET  20
Wiznet5500lwIP domeEthernet(ETHERNET_CS, SPI, ETHERNET_INT);
#define EthernetClient WiFiClient
uint32_t uidBuffer[4];  // Board unique ID
byte MAC_Address[6];    // Mac address, uses part of the unique ID
#define CMD_SERVER_PORT 2323
std::atomic<bool> ethernetPresent;
IPConfig ServerConfig;


#include <WiFi.h>
#define SHUTTER_PORT 2424
std::atomic<bool> wifiPresent = false;
WIFIConfig wifiConfig;

std::atomic<bool> core0Ready = false;

// function prototypes
void getMacAddress(byte* macBuffer, uint32_t* uid);
void configureEthernet();
bool initEthernet(bool bUseDHCP, IPAddress ip, IPAddress dns, IPAddress gateway, IPAddress subnet);
void PingEthernetGateway();
void configureWiFi();
bool initWiFi(IPAddress ip, String sSSID, String sPassword);
void PingWiFiShutter();


//
// Setup and main loops
//
void setup()
{
	core0Ready = false;

	// Ethernet IP config
	ServerConfig.bUseDHCP = true;
	// fallback if dhcp fails
	ServerConfig.ip.fromString("192.168.0.99");
	ServerConfig.dns.fromString("192.168.0.1");
	ServerConfig.gateway.fromString("192.168.0.1");
	ServerConfig.subnet.fromString("255.255.255.0");
	// WiFi AP IP config
	wifiConfig.ip.fromString("172.31.255.1");
	strncpy(wifiConfig.sSSID,"RTIShutter", WIFI_VAR_LEN);
	strncpy(wifiConfig.sPassword,"RTIShutter", WIFI_VAR_LEN);

#ifdef DEBUG
	DebugPort.begin(115200);
	delay(1000);
	DBPrintln("========== RTI-Zone controller booting ==========");
#endif

	domeEthernet.setSPISpeed(30000000);
	digitalWrite(ETHERNET_RESET, 0);
	pinMode(ETHERNET_RESET, OUTPUT);
	getMacAddress(MAC_Address, uidBuffer);
	DBPrintln("MAC : " + String(MAC_Address[0], HEX) + String(":") +
					String(MAC_Address[1], HEX) + String(":") +
					String(MAC_Address[2], HEX) + String(":") +
					String(MAC_Address[3], HEX) + String(":") +
					String(MAC_Address[4], HEX) + String(":") +
					String(MAC_Address[5], HEX) );
	configureEthernet();

	configureWiFi();

	core0Ready = true;
	DBPrintln("========== Core 0 ready ==========");
}

void setup1()
{
	while(!core0Ready)
		delay(100);
	DBPrintln("========== Core 1 ready ==========");
}

//
// This loop takes care of all communications and commands
//
void loop()
{

	DBPrintln("==========Ping ethernet gateway");
	if(ethernetPresent) {
		PingEthernetGateway();
	}
	else {
		configureEthernet();
	}

	DBPrintln("==========Ping wifi client");
	if(wifiPresent) {
		PingWiFiShutter();
	}
	else {
		configureWiFi();
	}

	delay(1000); // ping once a second
}

//
// This loop does all the motor controls
//
void loop1()
{   
	// in the real code this runs some code to control a stepper motor via AccelStepper
	delay(1000);
}

void configureEthernet()
{
        DBPrintln("========== Configuring Ethernet ==========");
        ethernetPresent =  initEthernet(ServerConfig.bUseDHCP,
										ServerConfig.ip,
										ServerConfig.dns,
										ServerConfig.gateway,
										ServerConfig.subnet);
}


bool initEthernet(bool bUseDHCP, IPAddress ip, IPAddress dns, IPAddress gateway, IPAddress subnet)
{
	bool bDhcpOk;
	int nTimeout = 0;
	DBPrintln("========== Init Ethernet ==========");
	resetChip(ETHERNET_RESET);
	domeEthernet.setHostname("RTI-Dome");

	DBPrintln("========== Setting IP config ==========");
	// try DHCP if set
	if(bUseDHCP) {
		domeEthernet.config(INADDR_NONE);
		bDhcpOk = domeEthernet.begin(MAC_Address);
		while (domeEthernet.status() == WL_DISCONNECTED) {
			DBPrintln("Waiting for DHCP");
			delay(1000);
			nTimeout++;
			if(nTimeout>15) { // 15 seconds should be plenty
				bDhcpOk = false;
				break;
			}
		}
		if(!bDhcpOk) {
			DBPrintln("DHCP Failed!");
			if(domeEthernet.linkStatus() == LinkON ) {
				domeEthernet.config(ip, dns, gateway, subnet);
				domeEthernet.begin(MAC_Address);
			}
			else {
				DBPrintln("No cable");
				return false;
			}
		}
	}
	else {
		domeEthernet.config(ip, dns, gateway, subnet);
		domeEthernet.begin(MAC_Address);
	}

	DBPrintln("========== Checking hardware status ==========");
	if(domeEthernet.status() == WL_NO_SHIELD) {
		 DBPrintln("NO HARDWARE !!!");
		return false;
	}
	DBPrintln("W5500 Ok.");

	DBPrintln("IP = " + IpAddress2String(domeEthernet.localIP()));
	return true;
}


void PingEthernetGateway()
{
	int nRes = 0;
	nRes = domeEthernet.ping(domeEthernet.gatewayIP(),2,25); // ttl of 2 is enough to ping the gateway. 25ms timeout is also plenty as the ping is under 2.5ms
	DBPrintln("Gateway (" + IpAddress2String(domeEthernet.gatewayIP()) +") ping result : " + String(nRes>0?"Ok":"Error"));
	if(nRes<0) {
		DBPrintln("Ethetnet disconnected");
		ethernetPresent = false;
	}
}


void configureWiFi()
{
	DBPrintln("========== Configuring WiFi ==========");
	wifiPresent = initWiFi(wifiConfig.ip,
								String(wifiConfig.sSSID),
								String(wifiConfig.sPassword));
}

bool initWiFi(IPAddress ip, String sSSID, String sPassword)
{
	WiFi.disconnect();
	WiFi.mode(WIFI_AP);
	WiFi.setHostname("RTI-Dome");
	WiFi.config(ip);
	WiFi.beginAP(sSSID.c_str(), sPassword.c_str());
	if(WiFi.status() != WL_CONNECTED) {
		DBPrintln("========== Failed to start WiFi AP ==========");
		return false;
	}
	DBPrintln("IP = " + IpAddress2String(WiFi.localIP()));
	WiFi.setTimeout(25);

	return true;
}


void PingWiFiShutter()
{
	int nRes = 0;
	nRes = WiFi.ping("172.31.255.2",2);
	DBPrintln("WiFi 172.31.255.2 ping result : " + String(nRes>0?"Ok":"Error"));
	if(nRes<0) {
		DBPrintln("WiFi client disconnected");
	}
}


// reset chip with /reset connected to nPin
void resetChip(int nPin)
{
	digitalWrite(nPin, 0);
	delay(2);
	digitalWrite(nPin, 1);
	delay(10);
}

//
// Create MAC address from unique ID
//


#include <hardware/flash.h>
void getMacAddress(byte* macBuffer, uint32_t* uid)
{
	uint8_t UniqueID[8];
	flash_get_unique_id(UniqueID);
	macBuffer[0] = 0x52;
	macBuffer[1] = 0x54;
	macBuffer[2] = 0x49;
	macBuffer[3] = (byte)(UniqueID[5]);
	macBuffer[4] = (byte)(UniqueID[6]);
	macBuffer[5] = (byte)(UniqueID[7]);
}



